@@ -67,7 +67,7 @@ ExecStop=/bin/kill -INT $MAINPID
 WantedBy=multi-user.target
 """
 
-__version__ = '2.0.10'
+__version__ = '2.0.11'
 __all__ = (
 	'AdvancedHTTPServer',
 	'RegisterPath',
@@ -107,6 +107,7 @@ import time
 import traceback
 import unittest
 import urllib
+import weakref
 import zlib
 
 if sys.version_info[0] < 3:
@@ -114,6 +115,7 @@ if sys.version_info[0] < 3:
 	import cgi as html
 	import Cookie
 	import httplib
+	import Queue as queue
 	import SocketServer as socketserver
 	import urlparse
 	http = type('http', (), {'client': httplib, 'cookies': Cookie, 'server': BaseHTTPServer})
@@ -127,6 +129,7 @@ else:
 	import http.client
 	import http.cookies
 	import http.server
+	import queue
 	import socketserver
 	import urllib.parse
 	from configparser import ConfigParser
@@ -389,6 +392,38 @@ def build_server_from_config(config, section_name, server_klass=None, handler_kl
 			server.serve_files_list_directories = config.getboolean(section_name, 'list_directories')
 	return server
 
+class _RequestEmbryo(object):
+	__slots__ = ('server', 'socket', 'address', 'created')
+	def __init__(self, server, client_socket, address, created=None):
+		server.request_embryos.append(self)
+		self.server = weakref.ref(server)
+		self.socket = client_socket
+		self.address = address
+		self.created = created or time.time()
+
+	def fileno(self):
+		return self.socket.fileno()
+
+	def serve_ready(self):
+		server = self.server()  # server is a weakref
+		if not server:
+			return False
+
+		try:
+			self.socket.do_handshake()
+		except ssl.SSLWantReadError:
+			return False
+		except (socket.error, OSError, ValueError):
+			self.socket.close()
+			server.request_embryos.remove(self)
+			return False
+
+		self.socket.settimeout(None)
+		server.request_embryos.remove(self)
+		server.request_queue.put((self.socket, self.address))
+		server.handle_request()
+		return True
+
 class RegisterPath(object):
 	"""
 	Register a path and handler with the global handler map. This can be
@@ -662,12 +697,55 @@ class ServerNonThreaded(http.server.HTTPServer, object):
 		if not hasattr(self, 'logger'):
 			self.logger = logging.getLogger('AdvancedHTTPServer')
 		self.allow_reuse_address = True
+		self.request_queue = queue.Queue()
+		self.request_embryos = []
 		self.using_ssl = False
 		super(ServerNonThreaded, self).__init__(*args, **kwargs)
 
+	def __repr__(self):
+		address = self.server_address[0]
+		if self.socket.family == socket.AF_INET:
+			address += ':' + str(self.server_address[1])
+		elif self.socket.family == socket.AF_INET6:
+			address = '[' + address + ']:' + str(self.server_address[1])
+		return "<{0} address: {1} ssl: {2!r}>".format(self.__class__.__name__, address, self.using_ssl)
+
+	@property
+	def read_checkable_fds(self):
+		return [self] + self.request_embryos
+
 	def get_config(self):
 		return self.__config
 
+	def get_request(self):
+		return self.request_queue.get(block=True, timeout=None)
+
+	def handle_request(self):
+		timeout = self.socket.gettimeout()
+		if timeout is None:
+			timeout = self.timeout
+		elif self.timeout is not None:
+			timeout = min(timeout, self.timeout)
+
+		try:
+			request, client_address = self.request_queue.get(block=True, timeout=timeout)
+		except queue.Empty:
+			return self.handle_timeout()
+		except OSError:
+			return
+
+		if self.verify_request(request, client_address):
+			try:
+				self.process_request(request, client_address)
+			except Exception:
+				self.handle_error(request, client_address)
+				self.shutdown_request(request)
+			except:
+				self.shutdown_request(request)
+				raise
+		else:
+			self.shutdown_request(request)
+
 	def finish_request(self, request, client_address):
 		try:
 			super(ServerNonThreaded, self).finish_request(request, client_address)
@@ -677,6 +755,17 @@ class ServerNonThreaded(http.server.HTTPServer, object):
 			self.logger.warning('KeyboardInterrupt encountered in finish_request')
 			self.shutdown()
 
+	def serve_ready(self):
+		client_socket, address = self.socket.accept()
+		if self.using_ssl:
+			client_socket.settimeout(0)
+			embryo = _RequestEmbryo(self, client_socket, address)
+			embryo.serve_ready()
+		else:
+			client_socket.settimeout(None)
+			self.request_queue.put((client_socket, address))
+			self.handle_request()
+
 	def server_bind(self, *args, **kwargs):
 		self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
 		super(ServerNonThreaded, self).server_bind(*args, **kwargs)
@@ -1044,10 +1133,7 @@ class RequestHandler(http.server.BaseHTTPRequestHandler, object):
 		if ext in self.extensions_map:
 			return self.extensions_map[ext]
 		ext = ext.lower()
-		if ext in self.extensions_map:
-			return self.extensions_map[ext]
-		else:
-			return self.extensions_map['']
+		return self.extensions_map[ext if ext in self.extensions_map else '']
 
 	def stock_handler_respond_unauthorized(self, query):
 		"""This method provides a handler suitable to be used in the handler_map."""
@@ -1173,7 +1259,7 @@ class RequestHandler(http.server.BaseHTTPRequestHandler, object):
 
 	def do_OPTIONS(self):
 		available_methods = list(x[3:] for x in dir(self) if x.startswith('do_'))
-		if 'RPC' in available_methods and len(self.rpc_handler_map) == 0:
+		if 'RPC' in available_methods and not self.rpc_handler_map:
 			available_methods.remove('RPC')
 		self.send_response(200)
 		self.send_header('Content-Length', 0)
@@ -1648,7 +1734,7 @@ class AdvancedHTTPServer(object):
 		"""
 		if addresses is None:
 			addresses = []
-		if address is None and len(addresses) == 0:
+		if address is None and not addresses:
 			if ssl_certfile is not None:
 				if os.getuid():
 					addresses.insert(0, ('0.0.0.0', 8443, True))
@@ -1709,7 +1795,7 @@ class AdvancedHTTPServer(object):
 			for server in self.sub_servers:
 				if not server.using_ssl:
 					continue
-				server.socket = self._ssl_ctx.wrap_socket(server.socket, server_side=True)
+				server.socket = self._ssl_ctx.wrap_socket(server.socket, server_side=True, do_handshake_on_connect=False)
 
 		if hasattr(handler_klass, 'custom_authentication'):
 			self.logger.debug('a custom authentication function is being used')
@@ -1748,6 +1834,15 @@ class AdvancedHTTPServer(object):
 	def server_started(self):
 		return self.__server_thread is not None
 
+	def _serve_ready(self):
+		read_check = [self.__wakeup_fd]
+		for sub_server in self.sub_servers:
+			read_check.extend(sub_server.read_checkable_fds)
+		all_read_ready, _, _ = select.select(read_check, [], [])
+		for read_ready in all_read_ready:
+			if isinstance(read_ready, (_RequestEmbryo, http.server.HTTPServer)):
+				read_ready.serve_ready()
+
 	def serve_forever(self, fork=False):
 		"""
 		Start handling requests. This method must be called and does not
@@ -1772,10 +1867,7 @@ class AdvancedHTTPServer(object):
 		self.__is_running.set()
 		while not self.__should_stop.is_set():
 			try:
-				read_ready, _, _ = select.select([self.__wakeup_fd] + self.sub_servers, [], [])
-				for server in read_ready:
-					if isinstance(server, http.server.HTTPServer):
-						server.handle_request()
+				self._serve_ready()
 			except socket.error:
 				self.logger.warning('encountered socket error, stopping server')
 				self.__should_stop.set()
@@ -1920,7 +2012,7 @@ class AdvancedHTTPServer(object):
 			self.logger.info('basic authentication has been enabled')
 		if pwtype != 'plain':
 			algorithms_available = getattr(hashlib, 'algorithms_available', ()) or getattr(hashlib, 'algorithms', ())
-			if not pwtype in algorithms_available:
+			if pwtype not in algorithms_available:
 				raise ValueError('hashlib does not support the desired algorithm')
 			# only md5 and sha1 hex for backwards compatibility
 			if pwtype == 'md5' and len(password) == 32:
